/**
 * Build project
 * @function build
 * @param {string} [dirname=process.cwd()] - Project directory name
 * @param {Object} [options={}] - Optional settings
 * @returns {Promise}
 */
'use strict'

const path = require('path')
const coz = require('coz')
const aglob = require('aglob')
const lebab = require('lebab')
const {readFileAsync, writeFileAsync, mkdirpAsync, statAsync} = require('asfs')
const awatch = require('awatch')
const theCode = require('the-code')
const abrowserify = require('abrowserify')
const webpack = require('webpack')
const asleep = require('asleep')
const argx = require('argx')
const filecopy = require('filecopy')
const ababel = require('ababel')
const fixpack = require('fixpack')
const babelify = require('babelify')
const rimraf = require('rimraf')

const _tmpl = (filename) => path.resolve(`${__dirname}/../assets/tmpl`, filename)

const {existsAsync} = require('asfs')

/** @lends build */
async function build (dirname = process.cwd(), options = {}) {
  const args = argx(arguments)
  options = args.pop('object') || {}
  dirname = args.shift('string') || process.cwd()
  const {
    budPattern = [
      '.*.bud',
      '+(bin|ci|doc|example|lib|misc|test)/**/.*.bud'
    ],
    libDir = 'lib',
    shimDir = 'shim',
    jsPattern = '**/+(*.jsx|*.js)',
    demoSrc = 'doc/demo/entrypoint.jsx',
    demoDest = 'doc/demo/bundle.js',
    watch = false,
    watchTriggerInterval = 100
  } = options

  const pkgPath = path.resolve(dirname, 'package.json')
  const esmShimDir = shimDir + '/esm'

  fixpack(pkgPath)
  const pkg = require(pkgPath)

  const plugins = [
    ['@babel/plugin-proposal-object-rest-spread'],
    ['@babel/plugin-proposal-optional-chaining'],
    ['@babel/plugin-proposal-class-properties'],
    ['@babel/plugin-transform-runtime', {}, 'the-script-build-runtime'],
  ]
  const buildBud = async () => coz.render(budPattern)
  const buildJs = async () => {
    await theCode().format(path.join(libDir, '**/*.*'), {ignore: ['**/index.*']})
    await ababel(jsPattern, {
      cwd: libDir,
      out: shimDir,
      presets: [
        ['@babel/preset-env', {}],
        ['@babel/preset-react', {}],
      ],
      plugins
    })

    // Generate esm shim
    {
      for (const filename of await aglob('**/+(*.js|*.jsx)', {cwd: libDir})) {
        const src = path.resolve(libDir, filename)
        const dest = path.resolve(esmShimDir, filename).replace(/\.jsx$/, '.js')
        const {code} = lebab.transform(
          String(await readFileAsync(src)),
          ['commonjs'],
        )
        await mkdirpAsync(path.dirname(dest))
        await writeFileAsync(dest, code)
      }
      await ababel(jsPattern, {
        cwd: esmShimDir,
        out: esmShimDir,
        presets: [
          ['@babel/preset-env', {
            modules: false,
          }],
          ['@babel/preset-react', {}],
        ],
        plugins
      })
    }

    // Copy json files
    {
      for (const filename of await aglob('**/*.json', {cwd: libDir})) {
        const src = path.resolve(libDir, filename)
        await filecopy(src, path.resolve(shimDir, filename), {mkdirp: true})
        await filecopy(src, path.resolve(esmShimDir, filename), {mkdirp: true})
      }
    }

    {
      const hasDefault = !!await statAsync(path.resolve(libDir, 'default.js')).catch(() => null)
      await filecopy(
        hasDefault ? _tmpl('indexWithDefault.mjs') : _tmpl('index.mjs'),
        path.resolve(shimDir, 'index.mjs'),
      )
    }
  }

  const buildDemo = async () => abrowserify(demoSrc, demoDest, {
    debug: true,
    transforms: [
      ['babelify', {
        presets: [
          ['@babel/preset-env'],
          ['@babel/preset-react'],
        ],
        plugins,
      }]
    ],
    requires: [
      [path.join(dirname, shimDir), {expose: pkg.name}]
    ]
  })

  const cwd = process.cwd()
  process.chdir(dirname)

  const demoExists = await existsAsync(demoSrc)

  await buildJs()
  await buildBud()
  if (demoExists) {
    await buildDemo()
  }

  if (watch) {
    let needsBuildJs = false
    let needsBuildBud = false
    await awatch(budPattern, () => { needsBuildBud = true })
    await awatch(jsPattern, () => { needsBuildJs = true }, {cwd: libDir})
    while (watch) {
      if (needsBuildBud) {
        needsBuildBud = false
        let nameToDecache = Object.keys(require.cache).filter((filename) => /bud$/.test(filename))
        for (let name of nameToDecache) {
          delete require.cache[name]
        }
        await buildBud()
      }
      if (needsBuildJs) {
        needsBuildJs = false
        await buildJs()
        if (demoExists) {
          await buildDemo()
        }
      }
      await asleep(watchTriggerInterval)
    }
  }

  process.chdir(cwd)
}

module.exports = build
